/*
*  msprior.c
*
* Copyright (C) 2006  Michael Hickerson and Naoki Takebayashi
*
* This file is a part of msprior.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of the
* License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
* 02110-1301, USA
*/


/*
* This program read in 3 positive integer values as the input.
*   msprior reps taxa loci
* reps is number of replications, taxa is the number of taxon pairs
* loci is number of loci.
* Additionally, it reads in a configuration file.
*
* Output: Each line is tab delimited, and the columns contain following info.
* 1. theta, from flat prior
* 2. tau, time when the pair got separated (with weird scaling)
*         Several classes of tau are chosen from flat prior, and tau for
*         a species get chosen from these classes.
* 3. migration rate, fixed value (=0).
* 4. recombination rate, from flat prior
* 5. integer index: 1 ... number of taxa
* 6. BottleTime
* 7. BottStr1
* 8. BottStr2
*
*/

/*
 Change Log
 * Fri May 12 2006 Naoki Takebayashi <ffnt@uaf.edu>
 - upper and lower bounds of prior distribution for theta is configurable
   uses gParam.upperTheta and gParam.lowerTheta
 - To make this to work, tauequlaizer uses gParam.upperTheta/2.  upperTheta
   needs to be passed to the stat program.  So the first element of the 
   output line of this program is upperTheta now.  msbayes.pl correctly
   receive this and pass this value to the option flag (-T) of the 
   stat program.

   * Tue Mar 14 2006 Naoki Takebayashi <ffnt@uaf.edu>
 - separated the function of reading in the config file, and setup parameters.
   This is now done in a separate file setup.c.  Setting up parameters
   are cleaner now.
 - As the consequence, the number of taxon-pairs are not arbitrary limited
 - got rid of unused variables, and general clean-up
 - BottStr1 and 2 were forced to be 0.01 with probabilty 0.1.  This created
   rather weird prior distributions for the strengths of bottleneck.
   This is removed now.
 - Compared the results of the previous and new implementations.  Other than
   this arbitrary preference to stronger bottleneck, the results are 
   completely identical with the same seed.
*/
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsl/gsl_rng.h>	/* for base rand num gen's */
#include <gsl/gsl_randist.h>	/* for gsl_ran_gamma */
#include "msprior.h"
#include "setup.h"

/* This has to be Global */
const gsl_rng *gBaseRand;	/* global rand number generator */

/* 
* global variables which stores the parameters (settings for the upper
* limist of prior dist'n etc
*/
runParameters gParam;		/* stores upper limits of prior dist'n etc */
mutParameterArray gMutParam;	/* stores mut model & # samples for 
				   each taxon pair */
constrainedParameterArray gConParam;	/* store constrained sub-parameters for
					   each taxon pair */


int
comp_nums (const void *doubleNum1, const void *doubleNum2)
{
  const double *num1 = doubleNum1;
  const double *num2 = doubleNum2;
  if (*num1 < *num2)
    return -1;
  else if (*num1 > *num2)
    return 1;
  else
    return 0;
}

int
main (int argc, char *argv[])
{
  double N1, N2, Nanc, *tauArray = NULL,*UnconstrainedTauArray = NULL, theta, tauequalizer, gaussTime,
    mig, rec, BottStr1, BottStr2, BottleTime;
  int tauClass, *PSIarray = NULL;
  unsigned int numTauClasses = -1, u, locus, zzz, c;
  unsigned long randSeed;
  unsigned long long rep;
  extern const gsl_rng *gBaseRand;
  void display_nums (int *, int);
  int comp_nums (const void *, const void *);
  FILE *fpTauPsiArray;

  int b_constrain = 0;
  int *subParamConstrainConfig = NULL;

  /* set up gParam and gMutParam, as well as gConParam if constrain */
  LoadConfiguration (argc, argv);

  /* set b_constrain to 1 if constrain */
  if (gParam.constrain > 0)
    {
      //initialize constrain indicator
      b_constrain = 1;


      //initialize subParamConstrainConfig array
      subParamConstrainConfig = calloc (NUMBER_OF_CONPARAM, sizeof (int));
      if (subParamConstrainConfig == NULL)
	{
	  fprintf (stderr,
		   "ERROR: Not enough memory for subParamConstrainConfig\n");
	  exit (EXIT_FAILURE);
	}


      int i = 0;

      for (i = 0; i < strlen (gParam.subParamConstrain); i++)
	{
	  char a = (gParam.subParamConstrain)[i];

	  if (a == '1')
	    subParamConstrainConfig[i] = 1;
	  else if (a == '0')
	    subParamConstrainConfig[i] = 0;
	}

    }

  /* for initiating the gsl random number generator */
  /* initialize PRNG */
  srand (gParam.prngSeed);	/* Better way of seeding here ? */
  randSeed = rand ();
  if (debug_level > 0)
    randSeed = 1;

  gBaseRand = gsl_rng_alloc (gsl_rng_mt19937);	/* set the base PRNG to
						   Mersenne Twister */
  gsl_rng_set (gBaseRand, randSeed);	/* seed the PRNG */

  if (b_constrain == 0)
    {
      /* fixed numTauClasses configuration */
      if (gParam.numTauClasses != 0)
	{
	  if (gParam.numTauClasses > gParam.numTaxaPair)
	    {
	      fprintf (stderr, "WARN: numTauClasses (%u) is larger than "
		       "numTaxaPair (%u). Setting numTauClasses to %u",
		       gParam.numTauClasses, gParam.numTaxaPair,
		       gParam.numTaxaPair);
	      gParam.numTauClasses = gParam.numTaxaPair;
	    }
	  numTauClasses = gParam.numTauClasses;
	}
    }
  else
    {
      numTauClasses = gParam.numTauClasses;
    }

  /* Sizes are set to the number of taxon pairs (Max number of tau) */
  tauArray = calloc (gParam.numTaxaPair, sizeof (double));
  if (tauArray == NULL)
    {
      fprintf (stderr, "ERROR: Not enough memory for tauArray\n");
      exit (EXIT_FAILURE);
    }

  UnconstrainedTauArray = calloc (gParam.numTaxaPair, sizeof (double));
  if (UnconstrainedTauArray == NULL)
    {
      fprintf (stderr, "ERROR: Not enough memory for UnconstrainedTauArray\n");
      exit (EXIT_FAILURE);
    }

  PSIarray = calloc (gParam.numTaxaPair, sizeof (int));
  if (PSIarray == NULL)
    {
      fprintf (stderr, "ERROR: Not enough memory for PSIarray\n");
      exit (EXIT_FAILURE);
    }

  /* open the stream to output tauArray and PSIarray */
  if ((fpTauPsiArray = fopen (gParam.priorOutFile, "w")) == NULL)
    {
      fprintf (stderr, "Cannot open the file.\n");
      exit (1);
    }

  /* print out the column headers to the tauArray and PSIarray  */
  /* PRI.numTauClass PRI.Tau.1  PRI.Tau.2 PRI.Tau.3 ... PRI.Psi.1 PRI.Psi.2 PRI.Psi.3 ... */
  fprintf (fpTauPsiArray, "PRI.numTauClass");
  if (gParam.numTauClasses > 0)
    {				/* constrained psi analysis */
      for (zzz = 0; zzz < numTauClasses; zzz++)
	fprintf (fpTauPsiArray, "\tPRI.Tau.%d", zzz + 1);
      for (zzz = 0; zzz < numTauClasses; zzz++)
	fprintf (fpTauPsiArray, "\tPRI.Psi.%d", zzz + 1);
    }
  fprintf (fpTauPsiArray, "\n");


  /* Beginning of the main loop */
  for (rep = 0; rep < gParam.reps; rep++)
    {


      /*
       * Each taxon pair was separated at a time tau in the past.  Of
       * all pairs, some of them may have been separated at the same
       * time.  numTauClasses is the number of classes with different
       * divergence time.
       *
       * If gParam.numTauClasses is not set, we are sampling
       * numTauClasses from a uniform prior dist'n.
       */
      if (gParam.numTauClasses == 0)
	{			/* numTauClasses is NOT fixed */
	  numTauClasses =
	    1 + gsl_rng_uniform_int (gBaseRand, gParam.numTaxaPair);
	}

      for (c = 0; c < numTauClasses; c++)
	PSIarray[c] = 0;	/* Reset the PSIarray counters */

      int psiIndex = 0;
      // Randomly generate TauArray only when NOT constrain
      if ((b_constrain == 0) || (subParamConstrainConfig[0] != 1))
	{
	  /* sample tau's from uniform prior dist'n */
	  for (u = 0; u < numTauClasses; u++)
	    {
	      tauArray[u] = gsl_ran_flat (gBaseRand, 0.0, gParam.upperTau);
              if (debug_level)
		{
		  fprintf (stderr, "DEBUG:%u of %u categories:\t%lf\n",
			   u, numTauClasses, tauArray[u]);
		}
	    }//
           qsort (tauArray, numTauClasses, sizeof(double), comp_nums);
	}
      else if ((b_constrain == 1) && (subParamConstrainConfig[0] == 1))
	{

	  for (u = 0; u < numTauClasses; u++)
	    tauArray[u] = (gConParam.conData[u]).conTau;

	  double *conTauArray = NULL;

	  conTauArray = calloc (gParam.numTaxaPair, sizeof (double));
	  if (conTauArray == NULL)
	    {
	      fprintf (stderr, "ERROR: Not enough memory for conTauArray\n");
	      exit (EXIT_FAILURE);
	    }

	  for (u = 0; u < gParam.numTaxaPair; u++)
	    {
	      conTauArray[u] = gConParam.conData[u].conTau;
	    }

	  qsort (conTauArray, (gParam.numTaxaPair), sizeof (double),
		 comp_nums);


	  //initialize PSIarray
	  double tempTau = conTauArray[0];

	  PSIarray[psiIndex] = 1;
	  for (u = 1; u < gParam.numTaxaPair; u++)
	    {

	      if (conTauArray[u] == tempTau)
		PSIarray[psiIndex]++;
	      else
		{
		  tempTau = conTauArray[u];
		  PSIarray[++psiIndex]++;
		}
	    }

	  free (conTauArray);
	}

      //if(psiIndex != (numTauClasses-1) )
      // fprintf(stderr,"numberPsiArray:%d and numTauClasses:%d have problems\n", psiIndex, numTauClasses );

      
      int tauPsiIndex = 0;
      int tauCounter = 1;
      if ((b_constrain == 0) || (subParamConstrainConfig[0] != 1)) // not constrained
      { 
	int cc;
        for(cc = 0;cc < gParam.numTaxaPair; cc++)
	  UnconstrainedTauArray[cc] = 0.0;

	int counter = 0;
        while(1)
	  {
	      if(counter == gParam.numTaxaPair)
		break;
             
              if (counter < numTauClasses)  //make sure all Taus are covered
                tauClass = counter;
              else
                tauClass = gsl_rng_uniform_int (gBaseRand, numTauClasses);

              do{
                 int taxonIndex = gsl_rng_uniform_int (gBaseRand, gParam.numTaxaPair);
                 if(UnconstrainedTauArray[taxonIndex] == 0)
                   {
                     UnconstrainedTauArray[taxonIndex] = tauArray[tauClass];
                     PSIarray[tauClass] = PSIarray[tauClass] + 1;
                     break;
                   }// if
              }while(1);
             
              counter ++;
	  }//while(1)

      }// if
                   

      tauPsiIndex = 0;
      tauCounter = 1;
      for (u = 0; u < gParam.numTaxaPair; u++)
	{


	  //Check upperAncPopSize before doing anything
	  /* ancestral population size prior */
	  if (gParam.upperAncPopSize < 0.01)
	    {
	      fprintf (stderr,
		       "The upper bound (%lf) of ancestral pop. size is "
		       "smaller than the lower bound (0.01)\n",
		       gParam.upperAncPopSize);
	      exit (EXIT_FAILURE);
	    }// if

	  mutParameter taxonPairDat;

	  /* sample sizes, mutational model for u-th taxon-pair */
	  taxonPairDat = gMutParam.data[u];

	  constrainedParameter conTaxonPairDat;


	  BottStr1 = gsl_ran_flat (gBaseRand, 0.01, 1.0);
	  BottStr2 = gsl_ran_flat (gBaseRand, 0.01, 1.0);
	  /* timing of bottle neck */
	  BottleTime = gsl_ran_flat (gBaseRand, 0.000001, 1.0);

	  /* migration rate prior */
	  mig = gsl_ran_flat (gBaseRand, 0.0, gParam.upperMig);
	  /* theta prior */
	  theta = gsl_ran_flat (gBaseRand, gParam.lowerTheta,
				gParam.upperTheta);

	  /* population sizes immediately after the separation, and 
	     what it grows to after the bottleneck (today) */
	  N1 = gsl_ran_flat (gBaseRand, 0.01, 1.99);

	  /* Wen commented out the following in version 41, is it ok? */
	  /* ancestral population size prior */
	  //if(gParam.upperAncPopSize < 0.01) {
	  // fprintf(stderr, "The upper bound (%lf) of ancestral pop. size is "
	  //        "smaller than the lower bound (0.01)\n",
	  //        gParam.upperAncPopSize);
	  //exit(EXIT_FAILURE);
	  //}


	  /*
	     Nmax=((gParam.upperAncPopSize*gParam.upperTheta)*gParam.upperTheta)
	     /theta; 
	     Nanc = gsl_ran_flat (gBaseRand, 0.01, Nmax);
	   */

	  /* The upper limit of ancestral theta is defined by the product
	     of upper Theta (e.g. 40) and upper AncPopSize (e.g. 0.5) */
	  Nanc = gsl_ran_flat (gBaseRand, 0.01,
			       gParam.upperAncPopSize * gParam.upperTheta);

	  /* recombination rate */
	  rec = gsl_ran_flat (gBaseRand, 0.0, gParam.upperRec);

	  /* sample sizes, constrain model for u-th taxon-pair */
	  //conTaxonPairDat = gConParam.conData[u];

	  if (b_constrain == 1)
	    {

	      conTaxonPairDat = gConParam.conData[u];


	      /** bottleneck priors **/
	      /* severity of bottle neck (how small the population become) */
	      if (subParamConstrainConfig[1] == 1)
		BottStr1 = conTaxonPairDat.conBottPop1;
	      if (subParamConstrainConfig[2] == 1)
		BottStr2 = conTaxonPairDat.conBottPop2;
	      /* timing of bottle neck */
	      if (subParamConstrainConfig[3] == 1)
		BottleTime = conTaxonPairDat.conBottleTime;

	      /* migration rate prior */
	      if (subParamConstrainConfig[4] == 1)
		mig = conTaxonPairDat.conMig;

	      /* theta prior */
	      if (subParamConstrainConfig[5] == 1)
		theta = conTaxonPairDat.conTheta;

	      /* population sizes immediately after the separation, and 
	         what it grows to after the bottleneck (today) */
	      if (subParamConstrainConfig[6] == 1)
		N1 = conTaxonPairDat.conN1;

	      /* The upper limit of ancestral theta is defined by the product
	         of upper Theta (e.g. 40) and upper AncPopSize (e.g. 0.5) */
	      if (subParamConstrainConfig[7] == 1)
		Nanc = conTaxonPairDat.conNanc;

	      /* recombination rate */
	      if (subParamConstrainConfig[8] == 1)
		rec = conTaxonPairDat.conRec;
	    }

	  N2 = 2.0 - N1;
	  Nanc = Nanc / theta;	/* get the ratio of theta_anc / theta_cur 
				   This ratio is required for msDQH */

	  tauequalizer = gParam.upperTheta / 2 / theta;

	  /* pick a tau for every taxon-pair with replacement from the
	     array of X taxon-pairs, where X is a uniform discrete RV
	     from 1 to number of taxon-pairs */

	  if ((b_constrain == 0) || (subParamConstrainConfig[0] != 1))
	    {
              /*
	      // 1-31-2007 it will use new way of picking tau
	      if (u < numTauClasses)
		tauClass = u;
	      else
		tauClass = gsl_rng_uniform_int (gBaseRand, numTauClasses);
	      */

	      gaussTime = UnconstrainedTauArray[u];

	      //printf("picking index: %d, gaussTime: %lf  ", tauClass, tauArray[tauClass]);
	    } // if
	  else if ((b_constrain == 1) && (subParamConstrainConfig[0] == 1))
	    {
	      if (u < numTauClasses)
		{
		  tauClass = u;
		} // if 
	      else
		{
		  if (tauCounter <= ((PSIarray[tauPsiIndex]) - 1))
		    {
		      tauClass = tauPsiIndex;
		      tauCounter++;

		    }// if
		  else
		    {
		      tauCounter = 1;
		      int x = 1;
		      for (x = 1; x < numTauClasses; x++)
			if (gConParam.conData[u].conTau == tauArray[x])
			  tauPsiIndex = x;
		      //tauPsiIndex ++;
		      tauClass = tauPsiIndex;
		      tauCounter++;

		    } // if else
		} // if else

	      if (tauClass < numTauClasses)
		gaussTime = tauArray[tauClass];
	      else
		fprintf (stderr, "what happened to tauClass?\n");
	    }// if else
	  //gaussTime = tauArray[u];
	  //fprintf(stderr, "gaussTime")

	  /* use the following if simulating a particular fixed history */
	  /* gaussTime = tauArray[u]; */

	  gaussTime = gaussTime * tauequalizer;

	  /* The following 2 if's are weird */
	  if (gaussTime < 0.0001)
	    gaussTime = 0.0001;

	  BottleTime = BottleTime * 0.95 * gaussTime;

	  if (gaussTime < 0.0001)
	    BottleTime = 0.00005;

	  if (debug_level)
	    fprintf (stderr, "DEBUG: BottleTime:%lf\tgaussTime:%lf\n",
		     BottleTime, gaussTime);

	  /* print out the results */
	  for (locus = 0; locus < gParam.numLoci; locus++)
	    {

	      printf ("%lf %lf %lf %lf %lf %u ",
		      gParam.upperTheta, theta, gaussTime, mig, rec, u + 1);
	      printf ("%lf %lf %lf ", BottleTime, BottStr1, BottStr2);
	      printf ("%u %u %u %lf %lf %lf ",
		      taxonPairDat.numPerTaxa,
		      taxonPairDat.sample[0], taxonPairDat.sample[1],
		      taxonPairDat.tstv[0], taxonPairDat.tstv[1],
		      taxonPairDat.gamma);
	      printf ("%u %u %lf %lf %lf ",
		      taxonPairDat.seqLen, numTauClasses, N1, N2, Nanc);
	      printf ("%lf %lf %lf %lf ",
		      taxonPairDat.freqA, taxonPairDat.freqC,
		      taxonPairDat.freqG, taxonPairDat.freqT);

	      printf ("%u\n", gParam.numTaxaPair);

	      /* These feed into the system command line (msDQH) within
	         the perl shell msbayes.  Some of these are used directly
	         by msDQH, but some are also passed on to the sumstats
	         programs via the msDQH commabnd line, .... like bp[u],
	         theta, gaussTime, NumPerTax[u], yy, */
	    }// for
	}

      fprintf (fpTauPsiArray, "%d", gParam.numTauClasses);
      if (gParam.numTauClasses > 0)
	{			/* constrained psi analysis */
	  for (zzz = 0; zzz < numTauClasses; zzz++)
	    fprintf (fpTauPsiArray, "\t%lf", tauArray[zzz]);
	  for (zzz = 0; zzz < numTauClasses; zzz++)
	    fprintf (fpTauPsiArray, "\t%d", PSIarray[zzz]);
	}
      fprintf (fpTauPsiArray, "\n");
    }

  fclose (fpTauPsiArray);
  free (tauArray);
  free (UnconstrainedTauArray);
  free (PSIarray);
  free (subParamConstrainConfig);
  exit (0);
}
